# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

from pathlib import Path
import tarfile

from pygit2 import Index, Oid, Tree, Object


TREE_HASH = 'fd937514cb799514d4b81bb24c5fcfeb6472b245'
COMMIT_HASH = '2be5719152d4f82c7302b1c0932d8e5f0a4a0e98'


def check_writing(repo, treeish, timestamp=None):
    archive = tarfile.open('foo.tar', mode='w')
    repo.write_archive(treeish, archive)

    index = Index()
    if isinstance(treeish, Object):
        index.read_tree(treeish.peel(Tree))
    else:
        index.read_tree(repo[treeish].peel(Tree))

    assert len(index) == len(archive.getmembers())

    if timestamp:
        fileinfo = archive.getmembers()[0]
        assert timestamp == fileinfo.mtime

    archive.close()
    path = Path('foo.tar')
    assert path.is_file()
    path.unlink()


def test_write_tree(testrepo):
    check_writing(testrepo, TREE_HASH)
    check_writing(testrepo, Oid(hex=TREE_HASH))
    check_writing(testrepo, testrepo[TREE_HASH])


def test_write_commit(testrepo):
    commit_timestamp = testrepo[COMMIT_HASH].committer.time
    check_writing(testrepo, COMMIT_HASH, commit_timestamp)
    check_writing(testrepo, Oid(hex=COMMIT_HASH), commit_timestamp)
    check_writing(testrepo, testrepo[COMMIT_HASH], commit_timestamp)



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for Odb backends."""

# Standard Library
import binascii
from pathlib import Path

import pytest

# pygit2
import pygit2
from pygit2.enums import ObjectType
from . import utils


BLOB_HEX = 'af431f20fc541ed6d5afede3e2dc7160f6f01f16'
BLOB_RAW = binascii.unhexlify(BLOB_HEX.encode('ascii'))
BLOB_OID = pygit2.Oid(raw=BLOB_RAW)


@pytest.fixture
def odb(barerepo):
    yield barerepo.odb, Path(barerepo.path) / 'objects'


def test_pack(odb):
    odb, path = odb

    pack = pygit2.OdbBackendPack(path)
    assert len(list(pack)) > 0
    for obj in pack:
        assert obj in odb


def test_loose(odb):
    odb, path = odb

    pack = pygit2.OdbBackendLoose(path, 5, False)
    assert len(list(pack)) > 0
    for obj in pack:
        assert obj in odb


class ProxyBackend(pygit2.OdbBackend):
    def __init__(self, source):
        super().__init__()
        self.source = source

    def read_cb(self, oid):
        return self.source.read(oid)

    def read_prefix_cb(self, oid):
        return self.source.read_prefix(oid)

    def read_header_cb(self, oid):
        typ, data = self.source.read(oid)
        return typ, len(data)

    def exists_cb(self, oid):
        return self.source.exists(oid)

    def exists_prefix_cb(self, oid):
        return self.source.exists_prefix(oid)

    def refresh_cb(self):
        self.source.refresh()

    def __iter__(self):
        return iter(self.source)


#
# Test a custom object backend alone (without adding it to an ODB)
# This doesn't make much sense, but it's possible.
#


@pytest.fixture
def proxy(barerepo):
    path = Path(barerepo.path) / 'objects'
    yield ProxyBackend(pygit2.OdbBackendPack(path))


def test_iterable(proxy):
    assert BLOB_HEX in [o for o in proxy]


def test_read(proxy):
    with pytest.raises(TypeError):
        proxy.read(123)
    utils.assertRaisesWithArg(KeyError, '1' * 40, proxy.read, '1' * 40)

    ab = proxy.read(BLOB_OID)
    a = proxy.read(BLOB_HEX)
    assert ab == a
    assert (ObjectType.BLOB, b'a contents\n') == a


def test_read_prefix(proxy):
    a_hex_prefix = BLOB_HEX[:4]
    a3 = proxy.read_prefix(a_hex_prefix)
    assert (ObjectType.BLOB, b'a contents\n', BLOB_OID) == a3


def test_exists(proxy):
    with pytest.raises(TypeError):
        proxy.exists(123)

    assert not proxy.exists('1' * 40)
    assert proxy.exists(BLOB_HEX)


def test_exists_prefix(proxy):
    a_hex_prefix = BLOB_HEX[:4]
    assert BLOB_HEX == proxy.exists_prefix(a_hex_prefix)


#
# Test a custom object backend, through a Repository.
#


@pytest.fixture
def repo(barerepo):
    odb = pygit2.Odb()

    path = Path(barerepo.path) / 'objects'
    backend = pygit2.OdbBackendPack(path)
    backend = ProxyBackend(backend)
    odb.add_backend(backend, 1)

    repo = pygit2.Repository()
    repo.set_odb(odb)
    yield repo


def test_repo_read(repo):
    with pytest.raises(TypeError):
        repo[123]

    utils.assertRaisesWithArg(KeyError, '1' * 40, repo.__getitem__, '1' * 40)

    ab = repo[BLOB_OID]
    a = repo[BLOB_HEX]
    assert ab == a



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for Commit objects."""

import sys

import pytest

from pygit2 import Signature, Oid, GitError
from pygit2.enums import ObjectType
from . import utils


COMMIT_SHA = '5fe808e8953c12735680c257f56600cb0de44b10'
COMMIT_SHA_TO_AMEND = (
    '784855caf26449a1914d2cf62d12b9374d76ae78'  # tip of the master branch
)


@utils.refcount
def test_commit_refcount(barerepo):
    commit = barerepo[COMMIT_SHA]
    start = sys.getrefcount(commit)
    tree = commit.tree
    del tree
    end = sys.getrefcount(commit)
    assert start == end


def test_read_commit(barerepo):
    commit = barerepo[COMMIT_SHA]
    assert COMMIT_SHA == commit.id
    parents = commit.parents
    assert 1 == len(parents)
    assert 'c2792cfa289ae6321ecf2cd5806c2194b0fd070c' == parents[0].id
    assert commit.message_encoding is None
    assert commit.message == (
        'Second test data commit.\n\n' 'This commit has some additional text.\n'
    )
    commit_time = 1288481576
    assert commit_time == commit.commit_time
    assert commit.committer == Signature(
        'Dave Borowitz', 'dborowitz@google.com', commit_time, -420
    )
    assert commit.author == Signature(
        'Dave Borowitz', 'dborowitz@google.com', 1288477363, -420
    )
    assert '967fce8df97cc71722d3c2a5930ef3e6f1d27b12' == commit.tree.id


def test_new_commit(barerepo):
    repo = barerepo
    message = 'New commit.\n\nMessage with non-ascii chars: ééé.\n'
    committer = Signature('John Doe', 'jdoe@example.com', 12346, 0)
    author = Signature(
        'J. David Ibáñez', 'jdavid@example.com', 12345, 0, encoding='utf-8'
    )
    tree = '967fce8df97cc71722d3c2a5930ef3e6f1d27b12'
    tree_prefix = tree[:5]
    too_short_prefix = tree[:3]

    parents = [COMMIT_SHA[:5]]
    with pytest.raises(ValueError):
        repo.create_commit(None, author, committer, message, too_short_prefix, parents)

    sha = repo.create_commit(None, author, committer, message, tree_prefix, parents)
    commit = repo[sha]

    assert ObjectType.COMMIT == commit.type
    assert '98286caaab3f1fde5bf52c8369b2b0423bad743b' == commit.id
    assert commit.message_encoding is None
    assert message == commit.message
    assert 12346 == commit.commit_time
    assert committer == commit.committer
    assert author == commit.author
    assert tree == commit.tree.id
    assert Oid(hex=tree) == commit.tree_id
    assert 1 == len(commit.parents)
    assert COMMIT_SHA == commit.parents[0].id
    assert Oid(hex=COMMIT_SHA) == commit.parent_ids[0]


def test_new_commit_encoding(barerepo):
    repo = barerepo
    encoding = 'iso-8859-1'
    message = 'New commit.\n\nMessage with non-ascii chars: ééé.\n'
    committer = Signature('John Doe', 'jdoe@example.com', 12346, 0, encoding)
    author = Signature('J. David Ibáñez', 'jdavid@example.com', 12345, 0, encoding)
    tree = '967fce8df97cc71722d3c2a5930ef3e6f1d27b12'
    tree_prefix = tree[:5]

    parents = [COMMIT_SHA[:5]]
    sha = repo.create_commit(
        None, author, committer, message, tree_prefix, parents, encoding
    )
    commit = repo[sha]

    assert ObjectType.COMMIT == commit.type
    assert 'iso-8859-1' == commit.message_encoding
    assert message.encode(encoding) == commit.raw_message
    assert 12346 == commit.commit_time
    assert committer == commit.committer
    assert author == commit.author
    assert tree == commit.tree.id
    assert Oid(hex=tree) == commit.tree_id
    assert 1 == len(commit.parents)
    assert COMMIT_SHA == commit.parents[0].id
    assert Oid(hex=COMMIT_SHA) == commit.parent_ids[0]


def test_modify_commit(barerepo):
    message = 'New commit.\n\nMessage.\n'
    committer = ('John Doe', 'jdoe@example.com', 12346)
    author = ('Jane Doe', 'jdoe2@example.com', 12345)

    commit = barerepo[COMMIT_SHA]

    with pytest.raises(AttributeError):
        setattr(commit, 'message', message)
    with pytest.raises(AttributeError):
        setattr(commit, 'committer', committer)
    with pytest.raises(AttributeError):
        setattr(commit, 'author', author)
    with pytest.raises(AttributeError):
        setattr(commit, 'tree', None)
    with pytest.raises(AttributeError):
        setattr(commit, 'parents', None)


def test_amend_commit_metadata(barerepo):
    repo = barerepo
    commit = repo[COMMIT_SHA_TO_AMEND]
    assert commit.id == repo.head.target

    encoding = 'iso-8859-1'
    amended_message = 'Amended commit message.\n\nMessage with non-ascii chars: ééé.\n'
    amended_author = Signature(
        'Jane Author', 'jane@example.com', 12345, 0, encoding=encoding
    )
    amended_committer = Signature(
        'John Committer', 'john@example.com', 12346, 0, encoding=encoding
    )

    amended_oid = repo.amend_commit(
        commit,
        'HEAD',
        message=amended_message,
        author=amended_author,
        committer=amended_committer,
        encoding=encoding,
    )
    amended_commit = repo[amended_oid]

    assert repo.head.target == amended_oid
    assert ObjectType.COMMIT == amended_commit.type
    assert amended_committer == amended_commit.committer
    assert amended_author == amended_commit.author
    assert amended_message.encode(encoding) == amended_commit.raw_message
    assert commit.author != amended_commit.author
    assert commit.committer != amended_commit.committer
    assert commit.tree == amended_commit.tree  # we didn't touch the tree


def test_amend_commit_tree(barerepo):
    repo = barerepo
    commit = repo[COMMIT_SHA_TO_AMEND]
    assert commit.id == repo.head.target

    tree = '967fce8df97cc71722d3c2a5930ef3e6f1d27b12'
    tree_prefix = tree[:5]

    amended_oid = repo.amend_commit(commit, 'HEAD', tree=tree_prefix)
    amended_commit = repo[amended_oid]

    assert repo.head.target == amended_oid
    assert ObjectType.COMMIT == amended_commit.type
    assert commit.message == amended_commit.message
    assert commit.author == amended_commit.author
    assert commit.committer == amended_commit.committer
    assert commit.tree_id != amended_commit.tree_id
    assert Oid(hex=tree) == amended_commit.tree_id


def test_amend_commit_not_tip_of_branch(barerepo):
    repo = barerepo

    # This commit isn't at the tip of the branch.
    commit = repo['5fe808e8953c12735680c257f56600cb0de44b10']
    assert commit.id != repo.head.target

    # Can't update HEAD to the rewritten commit because it's not the tip of the branch.
    with pytest.raises(GitError):
        repo.amend_commit(commit, 'HEAD', message="this won't work!")

    # We can still amend the commit if we don't try to update a ref.
    repo.amend_commit(commit, None, message='this will work')


def test_amend_commit_no_op(barerepo):
    repo = barerepo
    commit = repo[COMMIT_SHA_TO_AMEND]
    assert commit.id == repo.head.target

    amended_oid = repo.amend_commit(commit, None)
    assert amended_oid == commit.id


def test_amend_commit_argument_types(barerepo):
    repo = barerepo

    some_tree = repo['967fce8df97cc71722d3c2a5930ef3e6f1d27b12']
    commit = repo[COMMIT_SHA_TO_AMEND]
    alt_commit1 = Oid(hex=COMMIT_SHA_TO_AMEND)
    alt_commit2 = COMMIT_SHA_TO_AMEND
    alt_tree = some_tree
    alt_refname = (
        repo.head
    )  # try this one last, because it'll change the commit at the tip

    # Pass bad values/types for the commit
    with pytest.raises(ValueError):
        repo.amend_commit(None, None)
    with pytest.raises(TypeError):
        repo.amend_commit(some_tree, None)

    # Pass bad types for signatures
    with pytest.raises(TypeError):
        repo.amend_commit(commit, None, author='Toto')
    with pytest.raises(TypeError):
        repo.amend_commit(commit, None, committer='Toto')

    # Pass bad refnames
    with pytest.raises(ValueError):
        repo.amend_commit(commit, 'this-ref-doesnt-exist')
    with pytest.raises(TypeError):
        repo.amend_commit(commit, repo)

    # Pass bad trees
    with pytest.raises(ValueError):
        repo.amend_commit(commit, None, tree="can't parse this")
    with pytest.raises(KeyError):
        repo.amend_commit(commit, None, tree='baaaaad')

    # Pass an Oid for the commit
    amended_oid = repo.amend_commit(alt_commit1, None, message='Hello')
    amended_commit = repo[amended_oid]
    assert ObjectType.COMMIT == amended_commit.type
    assert amended_oid != COMMIT_SHA_TO_AMEND

    # Pass a str for the commit
    amended_oid = repo.amend_commit(alt_commit2, None, message='Hello', tree=alt_tree)
    amended_commit = repo[amended_oid]
    assert ObjectType.COMMIT == amended_commit.type
    assert amended_oid != COMMIT_SHA_TO_AMEND
    assert repo[COMMIT_SHA_TO_AMEND].tree != amended_commit.tree
    assert alt_tree.id == amended_commit.tree_id

    # Pass an actual reference object for refname
    # (Warning: the tip of the branch will be altered after this test!)
    amended_oid = repo.amend_commit(alt_commit2, alt_refname, message='Hello')
    amended_commit = repo[amended_oid]
    assert ObjectType.COMMIT == amended_commit.type
    assert amended_oid != COMMIT_SHA_TO_AMEND
    assert repo.head.target == amended_oid



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

# Import setuptools before distutils to avoid user warning
from setuptools import setup, Extension

from distutils.command.build import build
from distutils.command.sdist import sdist
from distutils import log
import os
from pathlib import Path
from subprocess import Popen, PIPE
import sys

# Import stuff from pygit2/_utils.py without loading the whole pygit2 package
sys.path.insert(0, 'pygit2')
from _build import __version__, get_libgit2_paths

del sys.path[0]


libgit2_bin, libgit2_kw = get_libgit2_paths()


class sdist_files_from_git(sdist):
    def get_file_list(self):
        popen = Popen(
            ['git', 'ls-files'], stdout=PIPE, stderr=PIPE, universal_newlines=True
        )
        stdoutdata, stderrdata = popen.communicate()
        if popen.returncode != 0:
            print(stderrdata)
            sys.exit()

        def exclude(line):
            for prefix in ['.', 'appveyor.yml', 'docs/', 'misc/']:
                if line.startswith(prefix):
                    return True
            return False

        for line in stdoutdata.splitlines():
            if not exclude(line):
                self.filelist.append(line)

        # Ok
        self.filelist.sort()
        self.filelist.remove_duplicates()
        self.write_manifest()


classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12',
    'Programming Language :: Python :: 3.13',
    'Programming Language :: Python :: Implementation :: PyPy',
    'Programming Language :: Python :: Implementation :: CPython',
    'Topic :: Software Development :: Version Control',
    'Typing :: Typed',
]

__dir__ = Path(__file__).parent
long_description = (__dir__ / 'README.md').read_text()

cmdclass = {
    'sdist': sdist_files_from_git,
}


# On Windows, we install the git2.dll too.
class BuildWithDLLs(build):
    def _get_dlls(self):
        # return a list of (FQ-in-name, relative-out-name) tuples.
        ret = []
        bld_ext = self.distribution.get_command_obj('build_ext')
        compiler_type = bld_ext.compiler.compiler_type
        libgit2_dlls = []
        if compiler_type == 'msvc':
            libgit2_dlls.append('git2.dll')
        elif compiler_type == 'mingw32':
            libgit2_dlls.append('libgit2.dll')
        look_dirs = [libgit2_bin] + os.getenv('PATH', '').split(os.pathsep)

        target = Path(self.build_lib).absolute() / 'pygit2'
        for dll in libgit2_dlls:
            for look in look_dirs:
                f = Path(look) / dll
                if f.is_file():
                    ret.append((f, target))
                    break
            else:
                log.warn(f'Could not find required DLL {dll} to include')
                log.debug(f'(looked in {look_dirs})')
        return ret

    def run(self):
        build.run(self)
        for s, d in self._get_dlls():
            self.copy_file(s, d)


# On Windows we package up the dlls with the plugin.
if os.name == 'nt':
    cmdclass['build'] = BuildWithDLLs

src = __dir__ / 'src'
pygit2_exts = [str(path) for path in sorted(src.iterdir()) if path.suffix == '.c']
ext_modules = [Extension('pygit2._pygit2', pygit2_exts, **libgit2_kw)]

setup(
    name='pygit2',
    description='Python bindings for libgit2.',
    keywords='git',
    version=__version__,
    classifiers=classifiers,
    license='GPLv2 with linking exception',
    maintainer='J. David Ibáñez',
    maintainer_email='jdavid.ibp@gmail.com',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=['pygit2'],
    package_data={'pygit2': ['decl/*.h', '*.pyi', 'py.typed']},
    zip_safe=False,
    cmdclass=cmdclass,
    cffi_modules=['pygit2/_run.py:ffi'],
    ext_modules=ext_modules,
    # Requirements
    python_requires='>=3.10',
    setup_requires=['cffi>=1.17.0'],
    install_requires=['cffi>=1.17.0'],
    # URLs
    url='https://github.com/libgit2/pygit2',
    project_urls={
        'Documentation': 'https://www.pygit2.org/',
        'Changelog': 'https://github.com/libgit2/pygit2/blob/master/CHANGELOG.md',
        'Funding': 'https://github.com/sponsors/jdavid',
    },
)



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for note objects."""

from pygit2 import Signature
import pytest


NOTE = ('6c8980ba963cad8b25a9bcaf68d4023ee57370d8', 'note message')

NOTES = [
    (
        'ab533997b80705767be3dae8cbb06a0740809f79',
        'First Note - HEAD\n',
        '784855caf26449a1914d2cf62d12b9374d76ae78',
    ),
    (
        'd879714d880671ed84f8aaed8b27fca23ba01f27',
        'Second Note - HEAD~1\n',
        'f5e5aa4e36ab0fe62ee1ccc6eb8f79b866863b87',
    ),
]


def test_create_note(barerepo):
    annotated_id = barerepo.revparse_single('HEAD~3').id
    author = committer = Signature('Foo bar', 'foo@bar.com', 12346, 0)
    note_id = barerepo.create_note(NOTE[1], author, committer, str(annotated_id))
    assert NOTE[0] == note_id

    # check the note blob
    assert NOTE[1].encode() == barerepo[note_id].data


def test_lookup_note(barerepo):
    annotated_id = str(barerepo.head.target)
    note = barerepo.lookup_note(annotated_id)
    assert NOTES[0][0] == note.id
    assert NOTES[0][1] == note.message


def test_remove_note(barerepo):
    head = barerepo.head
    note = barerepo.lookup_note(str(head.target))
    author = committer = Signature('Foo bar', 'foo@bar.com', 12346, 0)
    note.remove(author, committer)
    with pytest.raises(KeyError):
        barerepo.lookup_note(str(head.target))


def test_iterate_notes(barerepo):
    for i, note in enumerate(barerepo.notes()):
        assert NOTES[i] == (note.id, note.message, note.annotated_id)


def test_iterate_non_existing_ref(barerepo):
    with pytest.raises(KeyError):
        barerepo.notes('refs/notes/bad_ref')



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""Tests for Mailmap."""

from pygit2 import Mailmap


TEST_MAILMAP = """\
# Simple Comment line
<cto@company.xx>                       <cto@coompany.xx>
Some Dude <some@dude.xx>         nick1 <bugs@company.xx>
Other Author <other@author.xx>   nick2 <bugs@company.xx>
Other Author <other@author.xx>         <nick2@company.xx>
Phil Hill <phil@company.xx>  # Comment at end of line
<joseph@company.xx>             Joseph <bugs@company.xx>
Santa Claus <santa.claus@northpole.xx> <me@company.xx>
"""

TEST_ENTRIES = [
    (None, 'cto@company.xx', None, 'cto@coompany.xx'),
    ('Some Dude', 'some@dude.xx', 'nick1', 'bugs@company.xx'),
    ('Other Author', 'other@author.xx', 'nick2', 'bugs@company.xx'),
    ('Other Author', 'other@author.xx', None, 'nick2@company.xx'),
    ('Phil Hill', None, None, 'phil@company.xx'),
    (None, 'joseph@company.xx', 'Joseph', 'bugs@company.xx'),
    ('Santa Claus', 'santa.claus@northpole.xx', None, 'me@company.xx'),
]

TEST_RESOLVE = [
    ('Brad', 'cto@company.xx', 'Brad', 'cto@coompany.xx'),
    ('Brad L', 'cto@company.xx', 'Brad L', 'cto@coompany.xx'),
    ('Some Dude', 'some@dude.xx', 'nick1', 'bugs@company.xx'),
    ('Other Author', 'other@author.xx', 'nick2', 'bugs@company.xx'),
    ('nick3', 'bugs@company.xx', 'nick3', 'bugs@company.xx'),
    ('Other Author', 'other@author.xx', 'Some Garbage', 'nick2@company.xx'),
    ('Phil Hill', 'phil@company.xx', 'unknown', 'phil@company.xx'),
    ('Joseph', 'joseph@company.xx', 'Joseph', 'bugs@company.xx'),
    ('Santa Claus', 'santa.claus@northpole.xx', 'Clause', 'me@company.xx'),
    ('Charles', 'charles@charles.xx', 'Charles', 'charles@charles.xx'),
]


def test_empty():
    mailmap = Mailmap()

    for _, _, name, email in TEST_RESOLVE:
        assert mailmap.resolve(name, email) == (name, email)


def test_new():
    mailmap = Mailmap()

    # Add entries to the mailmap
    for entry in TEST_ENTRIES:
        mailmap.add_entry(*entry)

    for real_name, real_email, name, email in TEST_RESOLVE:
        assert mailmap.resolve(name, email) == (real_name, real_email)


def test_parsed():
    mailmap = Mailmap.from_buffer(TEST_MAILMAP)

    for real_name, real_email, name, email in TEST_RESOLVE:
        assert mailmap.resolve(name, email) == (real_name, real_email)


# TODO: Add a testcase which uses .mailmap in a repo



# Copyright 2010-2025 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

from __future__ import annotations
from typing import TYPE_CHECKING

# Import from pygit2
from ._pygit2 import Oid
from .callbacks import git_fetch_options, git_push_options, git_remote_callbacks
from .enums import FetchPrune
from .errors import check_error
from .ffi import ffi, C
from .refspec import Refspec
from . import utils
from .utils import maybe_string, to_bytes, strarray_to_strings, StrArray

# Need BaseRepository for type hints, but don't let it cause a circular dependency
if TYPE_CHECKING:
    from .repository import BaseRepository


class TransferProgress:
    """Progress downloading and indexing data during a fetch."""

    def __init__(self, tp):
        self.total_objects = tp.total_objects
        """Total number of objects to download"""

        self.indexed_objects = tp.indexed_objects
        """Objects which have been indexed"""

        self.received_objects = tp.received_objects
        """Objects which have been received up to now"""

        self.local_objects = tp.local_objects
        """Local objects which were used to fix the thin pack"""

        self.total_deltas = tp.total_deltas
        """Total number of deltas in the pack"""

        self.indexed_deltas = tp.indexed_deltas
        """Deltas which have been indexed"""

        self.received_bytes = tp.received_bytes
        """"Number of bytes received up to now"""


class Remote:
    def __init__(self, repo: BaseRepository, ptr):
        """The constructor is for internal use only."""
        self._repo = repo
        self._remote = ptr
        self._stored_exception = None

    def __del__(self):
        C.git_remote_free(self._remote)

    @property
    def name(self):
        """Name of the remote"""

        return maybe_string(C.git_remote_name(self._remote))

    @property
    def url(self):
        """Url of the remote"""

        return maybe_string(C.git_remote_url(self._remote))

    @property
    def push_url(self):
        """Push url of the remote"""

        return maybe_string(C.git_remote_pushurl(self._remote))

    def connect(self, callbacks=None, direction=C.GIT_DIRECTION_FETCH, proxy=None):
        """Connect to the remote.

        Parameters:

        proxy : None or True or str
            Proxy configuration. Can be one of:

            * `None` (the default) to disable proxy usage
            * `True` to enable automatic proxy detection
            * an url to a proxy (`http://proxy.example.org:3128/`)
        """
        proxy_opts = ffi.new('git_proxy_options *')
        C.git_proxy_options_init(proxy_opts, C.GIT_PROXY_OPTIONS_VERSION)
        self.__set_proxy(proxy_opts, proxy)
        with git_remote_callbacks(callbacks) as payload:
            err = C.git_remote_connect(
                self._remote, direction, payload.remote_callbacks, proxy_opts, ffi.NULL
            )
            payload.check_error(err)

    def fetch(
        self,
        refspecs=None,
        message=None,
        callbacks=None,
        prune: FetchPrune = FetchPrune.UNSPECIFIED,
        proxy=None,
        depth=0,
    ):
        """Perform a fetch against this remote. Returns a <TransferProgress>
        object.

        Parameters:

        prune : enums.FetchPrune
            * `UNSPECIFIED`: use the configuration from the repository.
            * `PRUNE`: remove any remote branch in the local repository
               that does not exist in the remote.
            * `NO_PRUNE`: always keep the remote branches

        proxy : None or True or str
            Proxy configuration. Can be one of:

            * `None` (the default) to disable proxy usage
            * `True` to enable automatic proxy detection
            * an url to a proxy (`http://proxy.example.org:3128/`)

        depth : int
            Number of commits from the tip of each remote branch history to fetch.

            If non-zero, the number of commits from the tip of each remote
            branch history to fetch. If zero, all history is fetched.
            The default is 0 (all history is fetched).
        """
        with git_fetch_options(callbacks) as payload:
            opts = payload.fetch_options
            opts.prune = prune
            opts.depth = depth
            self.__set_proxy(opts.proxy_opts, proxy)
            with StrArray(refspecs) as arr:
                err = C.git_remote_fetch(self._remote, arr.ptr, opts, to_bytes(message))
                payload.check_error(err)

        return TransferProgress(C.git_remote_stats(self._remote))

    def ls_remotes(self, callbacks=None, proxy=None):
        """
        Return a list of dicts that maps to `git_remote_head` from a
        `ls_remotes` call.

        Parameters:

        callbacks : Passed to connect()

        proxy : Passed to connect()
        """

        self.connect(callbacks=callbacks, proxy=proxy)

        refs = ffi.new('git_remote_head ***')
        refs_len = ffi.new('size_t *')

        err = C.git_remote_ls(refs, refs_len, self._remote)
        check_error(err)

        results = []
        for i in range(int(refs_len[0])):
            ref = refs[0][i]
            local = bool(ref.local)
            if local:
                loid = Oid(raw=bytes(ffi.buffer(ref.loid.id)[:]))
            else:
                loid = None

            remote = {
                'local': local,
                'loid': loid,
                'name': maybe_string(ref.name),
                'symref_target': maybe_string(ref.symref_target),
                'oid': Oid(raw=bytes(ffi.buffer(ref.oid.id)[:])),
            }

            results.append(remote)

        return results

    def prune(self, callbacks=None):
        """Perform a prune against this remote."""
        with git_remote_callbacks(callbacks) as payload:
            err = C.git_remote_prune(self._remote, payload.remote_callbacks)
            payload.check_error(err)

    @property
    def refspec_count(self):
        """Total number of refspecs in this remote"""

        return C.git_remote_refspec_count(self._remote)

    def get_refspec(self, n):
        """Return the <Refspec> object at the given position."""
        spec = C.git_remote_get_refspec(self._remote, n)
        return Refspec(self, spec)

    @property
    def fetch_refspecs(self):
        """Refspecs that will be used for fetching"""

        specs = ffi.new('git_strarray *')
        err = C.git_remote_get_fetch_refspecs(specs, self._remote)
        check_error(err)
        return strarray_to_strings(specs)

    @property
    def push_refspecs(self):
        """Refspecs that will be used for pushing"""

        specs = ffi.new('git_strarray *')
        err = C.git_remote_get_push_refspecs(specs, self._remote)
        check_error(err)
        return strarray_to_strings(specs)

    def push(self, specs, callbacks=None, proxy=None, push_options=None):
        """
        Push the given refspec to the remote. Raises ``GitError`` on protocol
        error or unpack failure.

        When the remote has a githook installed, that denies the reference this
        function will return successfully. Thus it is strongly recommended to
        install a callback, that implements
        :py:meth:`RemoteCallbacks.push_update_reference` and check the passed
        parameters for successfull operations.

        Parameters:

        specs : [str]
            Push refspecs to use.

        proxy : None or True or str
            Proxy configuration. Can be one of:

            * `None` (the default) to disable proxy usage
            * `True` to enable automatic proxy detection
            * an url to a proxy (`http://proxy.example.org:3128/`)

        push_options : [str]
            Push options to send to the server, which passes them to the
            pre-receive as well as the post-receive hook.
        """
        with git_push_options(callbacks) as payload:
            opts = payload.push_options
            self.__set_proxy(opts.proxy_opts, proxy)
            with StrArray(specs) as refspecs, StrArray(push_options) as pushopts:
                pushopts.assign_to(opts.remote_push_options)
                err = C.git_remote_push(self._remote, refspecs.ptr, opts)
                payload.check_error(err)

    def __set_proxy(self, proxy_opts, proxy):
        if proxy is None:
            proxy_opts.type = C.GIT_PROXY_NONE
        elif proxy is True:
            proxy_opts.type = C.GIT_PROXY_AUTO
        elif type(proxy) is str:
            proxy_opts.type = C.GIT_PROXY_SPECIFIED
            # Keep url in memory, otherwise memory is freed and bad things happen
            self.__url = ffi.new('char[]', to_bytes(proxy))
            proxy_opts.url = self.__url
        else:
            raise TypeError('Proxy must be None, True, or a string')


class RemoteCollection:
    """Collection of configured remotes

    You can use this class to look up and manage the remotes configured
    in a repository.  You can access repositories using index
    access. E.g. to look up the "origin" remote, you can use

    >>> repo.remotes["origin"]
    """

    def __init__(self, repo: BaseRepository):
        self._repo = repo

    def __len__(self):
        with utils.new_git_strarray() as names:
            err = C.git_remote_list(names, self._repo._repo)
            check_error(err)
            return names.count

    def __iter__(self):
        cremote = ffi.new('git_remote **')
        for name in self._ffi_names():
            err = C.git_remote_lookup(cremote, self._repo._repo, name)
            check_error(err)

            yield Remote(self._repo, cremote[0])

    def __getitem__(self, name):
        if isinstance(name, int):
            return list(self)[name]

        cremote = ffi.new('git_remote **')
        err = C.git_remote_lookup(cremote, self._repo._repo, to_bytes(name))
        check_error(err)

        return Remote(self._repo, cremote[0])

    def _ffi_names(self):
        with utils.new_git_strarray() as names:
            err = C.git_remote_list(names, self._repo._repo)
            check_error(err)
            for i in range(names.count):
                yield names.strings[i]

    def names(self):
        """An iterator over the names of the available remotes."""
        for name in self._ffi_names():
            yield maybe_string(name)

    def create(self, name, url, fetch=None):
        """Create a new remote with the given name and url. Returns a <Remote>
        object.

        If 'fetch' is provided, this fetch refspec will be used instead of the
        default.
        """
        cremote = ffi.new('git_remote **')

        name = to_bytes(name)
        url = to_bytes(url)
        if fetch:
            fetch = to_bytes(fetch)
            err = C.git_remote_create_with_fetchspec(
                cremote, self._repo._repo, name, url, fetch
            )
        else:
            err = C.git_remote_create(cremote, self._repo._repo, name, url)

        check_error(err)

        return Remote(self._repo, cremote[0])

    def create_anonymous(self, url):
        """Create a new anonymous (in-memory only) remote with the given URL.
        Returns a <Remote> object.
        """
        cremote = ffi.new('git_remote **')
        url = to_bytes(url)
        err = C.git_remote_create_anonymous(cremote, self._repo._repo, url)
        check_error(err)
        return Remote(self._repo, cremote[0])

    def rename(self, name, new_name):
        """Rename a remote in the configuration. The refspecs in standard
        format will be renamed.

        Returns a list of fetch refspecs (list of strings) which were not in
        the standard format and thus could not be remapped.
        """

        if not new_name:
            raise ValueError('Current remote name must be a non-empty string')

        if not new_name:
            raise ValueError('New remote name must be a non-empty string')

        problems = ffi.new('git_strarray *')
        err = C.git_remote_rename(
            problems, self._repo._repo, to_bytes(name), to_bytes(new_name)
        )
        check_error(err)
        return strarray_to_strings(problems)

    def delete(self, name):
        """Remove a remote from the configuration

        All remote-tracking branches and configuration settings for the remote will be removed.
        """
        err = C.git_remote_delete(self._repo._repo, to_bytes(name))
        check_error(err)

    def set_url(self, name, url):
        """Set the URL for a remote"""
        err = C.git_remote_set_url(self._repo._repo, to_bytes(name), to_bytes(url))
        check_error(err)

    def set_push_url(self, name, url):
        """Set the push-URL for a remote"""
        err = C.git_remote_set_pushurl(self._repo._repo, to_bytes(name), to_bytes(url))
        check_error(err)

    def add_fetch(self, name, refspec):
        """Add a fetch refspec (str) to the remote"""

        err = C.git_remote_add_fetch(
            self._repo._repo, to_bytes(name), to_bytes(refspec)
        )
        check_error(err)

    def add_push(self, name, refspec):
        """Add a push refspec (str) to the remote"""

        err = C.git_remote_add_push(self._repo._repo, to_bytes(name), to_bytes(refspec))
        check_error(err)
